#include "../../spcme.h"

#include "apu.h"
#include "dsp.h"

// u16 capBuf[MIXBUFSIZE * 2];

// Play buffer, left buffer is first MIXBUFSIZE * 2 u16's, right buffer is next
u16 playBuffer[MIXBUFSIZE * 2 * 2];
volatile int soundCursor;
bool paused = false;
extern "C" void IntrHandlerAsm();
volatile int mixesNeeded = 0;

void SetupSound() {
    soundCursor = 0;

    powerON(POWER_SOUND);
    SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);

    TIMER0_DATA = TIMER_FREQ(MIXRATE);
    TIMER0_CR = TIMER_DIV_1 | TIMER_ENABLE;

    TIMER1_DATA = 0x10000 - MIXBUFSIZE;
    TIMER1_CR = TIMER_CASCADE | TIMER_IRQ_REQ | TIMER_ENABLE;

    // Timing stuff
    TIMER2_DATA = 0;
    TIMER2_CR = TIMER_DIV_64 | TIMER_ENABLE;

    TIMER3_DATA = 0;
    TIMER3_CR = TIMER_CASCADE | TIMER_ENABLE;

    mixesNeeded = 0;

//    *((vu32*)0x04000510) = (u32)capBuf;
//    *((vu16*)0x04000514) = (MIXBUFSIZE * 2) >> 1;
//    *((vu16*)0x04000508) = 0x83;
}

void StopSound() {
    powerOFF(POWER_SOUND);
    SOUND_CR = 0;
    TIMER0_CR = 0;
    TIMER1_CR = 0;
}

inline void SendArm9Command(u32 command) {
    IPC_FIFO_SEND = command;
}

#ifdef ARM7SPC
#ifndef SNESDS
void LoadSpc(const u8 *spc) {
    ApuReset();
    DspReset();

// 0 - A, 1 - X, 2 - Y, 3 - RAMBASE, 4 - DP, 5 - PC (Adjusted into rambase)
// 6 - Cycles (bit 0 - C, bit 1 - v, bit 2 - h, bits 3+ cycles left)
// 7 - Optable
// 8 - NZ

    APU_STATE[0] = spc[0x27]<<24; // A
    APU_STATE[1] = spc[0x28]<<24; // X
    APU_STATE[2] = spc[0x29]<<24; // Y
    SetStateFromRawPSW(APU_STATE, spc[0x2A]);
    APU_SP = 0x100 | spc[0x2B]; // SP
    APU_STATE[5] = APU_STATE[3] + (spc[0x25] | (spc[0x26] << 8)); // PC

    for (int i=0; i<=0xffff; i++) APU_MEM[i] = spc[0x100 + i];
    for (int i=0; i<=0x7f; i++) {
        DSP_MEM[i] = spc[0x10100 + i];
    }
    for (int i=0; i<=0x3f; i++) APU_EXTRA_MEM[i] = spc[0x101c0 + i];

    ApuPrepareStateAfterReload();
    DspPrepareStateAfterReload();
}
#endif
#endif

extern "C" {
void InterruptHandler(void) {
    u32 flags = IF & IE;

#ifndef SNESDS
    if (flags & IRQ_FIFO_NOT_EMPTY) {
//        do {
            u32 command = IPC_FIFO_RECIEVE;

            // Command received from the ARM9, let's check what it is
            switch (command) {
            case ARM7COMMAND_START_SOUND:
                SetupSound();
                break;
            case ARM7COMMAND_STOP_SOUND:
                StopSound();
                SPC_IPC->arm7State = ARM7STATE_WAITING;
                break;
#ifdef ARM7SPC
            case ARM7COMMAND_LOAD_SPC:
                LoadSpc(SPC_IPC->rawSpc);
                SPC_IPC->arm7State = ARM7STATE_READY;
                break;
#endif
            }
//        } while (!(IPC_FIFO_CR & IPC_FIFO_RECV_EMPTY));
    }
#else
    if (flags & IRQ_FIFO_NOT_EMPTY) {
        do {
            u32 command = IPC_FIFO_RECIEVE;

            // Command received from the ARM9, let's check what it is
            switch (command & 0xff) {
            case 1:
                // Reset
                StopSound();

                ApuReset();
                DspReset();

                SetupSound();
                paused = false;
                break;
            case 2:
                // Pause/unpause
                if (!paused) {
                    StopSound();
                } else {
                    SetupSound();
                }
                paused = !paused;
                break;

            case 3:
                // Execute cpu
                u32 cycles = command >> 8;
                *((vu16*)0x27fffe0) = 0;
                ApuExecute(cycles);
                *((vu16*)0x27fffe0) = 1;
                break;
            }
        } while (!(IPC_FIFO_CR & IPC_FIFO_RECV_EMPTY));
        VBLANK_INTR_WAIT_FLAGS |= IRQ_FIFO_NOT_EMPTY;
        IF = IRQ_FIFO_NOT_EMPTY;
        return;
    }
#endif

    if (flags & IRQ_TIMER1) {
#ifndef SNESDS
#ifdef ARM7SPC
        long long begin = TIMER2_DATA + ((long long)TIMER3_DATA << 19);
#endif
#endif
        soundCursor = MIXBUFSIZE - soundCursor;

        // Left channel
        int channel = soundCursor == 0 ? 0 : 1;
        SCHANNEL_TIMER(channel) = SOUND_FREQ(MIXRATE);
        SCHANNEL_SOURCE(channel) = (uint32)&(playBuffer[MIXBUFSIZE - soundCursor]);
        SCHANNEL_LENGTH(channel) = (MIXBUFSIZE * 2) >> 2;
        SCHANNEL_REPEAT_POINT(channel) = 0;
        SCHANNEL_CR(channel) = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(0x7F) | SOUND_PAN(0) | SOUND_16BIT;

        // Right channel
        channel = soundCursor == 0 ? 2 : 3;
        SCHANNEL_TIMER(channel) = SOUND_FREQ(MIXRATE);
        SCHANNEL_SOURCE(channel) = (uint32)&(playBuffer[(MIXBUFSIZE - soundCursor) + (MIXBUFSIZE * 2)]);
        SCHANNEL_LENGTH(channel) = (MIXBUFSIZE * 2) >> 2;
        SCHANNEL_REPEAT_POINT(channel) = 0;
        SCHANNEL_CR(channel) = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(0x7F) | SOUND_PAN(0x7F) | SOUND_16BIT;

#ifndef SNESDS
#ifdef ARM7SPC
        const int cyclesToExecute = spcCyclesPerSec / (MIXRATE / MIXBUFSIZE);
        ApuExecute(cyclesToExecute * 21);
        DspMixSamplesStereo(MIXBUFSIZE, &playBuffer[soundCursor]);
        long long end = TIMER2_DATA + ((long long)TIMER3_DATA << 19);
        SPC_IPC->cpuTime += end - begin;
//        SPC_IPC->dspTime += (TIMER2_DATA + ((long long)TIMER3_DATA << 19)) - end;
#else
        SPC_IPC->curCommand = IPC_FIFO_CR;
        for (int i = 0; i < MIXBUFSIZE; i++) {
            playBuffer[soundCursor + i] = SPC_IPC->playBuffer[i];
            playBuffer[soundCursor + i + (MIXBUFSIZE * 2)] = SPC_IPC->playBuffer[i + MIXBUFSIZE];
        }
        if (IPC_FIFO_CR & IPC_FIFO_ERROR) IPC_FIFO_CR |= IPC_FIFO_SEND_CLEAR;
        SendArm9Command(ARM9COMMAND_UPDATE_APU);
#endif
#else
        mixesNeeded++;
#endif
    }

#ifndef SNESDS
#ifdef ARM7SPC
    if (flags & IRQ_VBLANK) {
        // Debug time data
        SPC_IPC->curTime += TIMER2_DATA | ((long long)TIMER3_DATA << 19);
        TIMER2_CR = 0;
        TIMER3_CR = 0;
        TIMER2_DATA = 0;
        TIMER2_CR = TIMER_DIV_64 | TIMER_ENABLE;
        TIMER3_DATA = 0;
        TIMER3_CR = TIMER_CASCADE | TIMER_ENABLE;
    }
#endif
#endif

    VBLANK_INTR_WAIT_FLAGS |= flags;
    IF = flags;
}
}

u16 readtouch(u32 command) {
    u16 result;

    SERIAL_CR = SERIAL_ENABLE | 0xA01;
    SERIAL_DATA = command;
    while (SERIAL_CR & SERIAL_BUSY);

    SERIAL_DATA = 0;
    while (SERIAL_CR & SERIAL_BUSY);
    result = SERIAL_DATA;

    SERIAL_CR = SERIAL_ENABLE | 0x201;
    SERIAL_DATA = 0;
    while (SERIAL_CR & SERIAL_BUSY);

    return ((result & 0x7F) << 5) | (SERIAL_DATA >> 3);
}

int main(int argc, char ** argv) {
    // Reset the clock if needed
    rtcReset();

    for (int i = 0; i < MIXBUFSIZE * 4; i++) {
        playBuffer[i] = 0;
    }

    // Set up the interrupt handler
    IME = 0;
    IRQ_HANDLER = &IntrHandlerAsm;

    IE = IRQ_TIMER1 | IRQ_FIFO_NOT_EMPTY | IRQ_VBLANK;

    DISP_SR = DISP_VBLANK_IRQ;

    IPC_FIFO_CR = IPC_FIFO_ENABLE | IPC_FIFO_SEND_CLEAR | IPC_FIFO_RECV_IRQ;

    IF = ~0;
    IME = 1;

#ifdef SNESDS
    *((vu16*)0x27fffe0) = 0;
    ApuReset();
    DspReset();

    SetupSound();

    for(;;) {
/*        if (mixesNeeded) {
            DspMixSamplesStereo(MIXBUFSIZE, &playBuffer[soundCursor]);
            mixesNeeded--;
*/
        swiWaitForVBlank();
        dspPreamp = *((vu16*)0x27ffff2);
        *((vu32*)0x27ffff4) = XKEYS;
 		*(vu16*)0x27fffec=readtouch(TSC_MEASURE_X);
       	*(vu16*)0x27fffee=readtouch(TSC_MEASURE_Y);
//        }
    }
#else
    SPC_IPC->arm7State = ARM7STATE_WAITING;

    for (;;) {
        swiWaitForVBlank();
#ifdef ARM7SPC
        dspPreamp = *((vu16*)0x27ffff2);
#endif
        *((vu32*)0x27ffff4) = XKEYS;
        rtcGetTime((u8*)0x27fff00);
		*(vu16*)0x27fffec=readtouch(TSC_MEASURE_X);
		*(vu16*)0x27fffee=readtouch(TSC_MEASURE_Y);
    }
#endif

    return 0;
}

//////////////////////////////////////////////////////////////////////
