#define PAGE_4K		(0b01011 << 1)
#define PAGE_8K		(0b01100 << 1)
#define PAGE_16K	(0b01101 << 1)
#define PAGE_32K	(0b01110 << 1)
#define PAGE_64K	(0b00111 << 1)
#define PAGE_128K	(0b10000 << 1)
#define PAGE_256K	(0b10001 << 1)
#define PAGE_512K	(0b10010 << 1)
#define PAGE_1M		(0b10011 << 1)
#define PAGE_2M		(0b10100 << 1)
#define PAGE_4M		(0b10101 << 1)
#define PAGE_8M		(0b10110 << 1)
#define PAGE_16M	(0b10111 << 1)
#define PAGE_32M	(0b11000 << 1)
#define PAGE_64M	(0b11001 << 1)
#define PAGE_128M	(0b11010 << 1)
#define PAGE_256M	(0b11011 << 1)
#define PAGE_512M	(0b11100 << 1)
#define PAGE_1G		(0b11101 << 1)
#define PAGE_2G		(0b11110 << 1)
#define PAGE_4G		(0b11111 << 1)

@---------------------------------------------------------------------------------
	.section ".init"
	.global _start
@---------------------------------------------------------------------------------
	.align	4
	.arm
@---------------------------------------------------------------------------------
_start:
@---------------------------------------------------------------------------------
	mov	r0, #0x04000000			@ IME = 0;
	add	r0, r0, #0x208
	strh	r0, [r0]

	bl initMem

	mrc p15,0,r0,c9,c1,0
	bic r0,r0,#0xff
	add r0,r0,#0x4000	@(DTCM end)
	msr cpsr_c,#0x13
	sub sp,r0,#0x40		@SVC (swi) stack
	msr cpsr_c,#0x12
	sub sp,r0,#0x80		@IRQ stack
	msr cpsr_c,#0x1f
	sub sp,r0,#0x180	@user stack
	
	ldr r0, =0x04000204
	mov r1, #0x4000
	strh r1, [r0]

	ldr	r1, =__itcm_lma		@ Copy instruction tightly coupled memory (itcm section) from LMA to VMA (ROM to RAM)
	ldr	r2, =__itcm_start
	ldr	r4, =__itcm_end
	bl	CopyMemCheck

	ldr	r1, =__dtcm_lma		@ Copy data tightly coupled memory (dtcm section) from LMA to VMA (ROM to RAM)
	ldr	r2, =__dtcm_start
	ldr	r4, =__dtcm_end
	bl	CopyMemCheck
/*
	ldr	r0, =__bss_start	@ Clear BSS section
	ldr	r1, =__bss_end
	sub	r1, r1, r0
	bl	ClearMem

	ldr	r0, =__sbss_start	@ Clear SBSS section 
	ldr	r1, =__sbss_end
	sub	r1, r1, r0
	bl	ClearMem
*/
	ldr	r1, =fake_heap_end	@ set heap end
	ldr	r0, =__eheap_end
	str	r0, [r1]
	
	ldr	r3, =_init		@ global constructors
	bl	_call_via_r3

	mov	r0, #0			@ int argc
	mov	r1, #0			@ char *argv[]
	ldr	r3, =main
	bl	_call_via_r3		@ jump to user code
		
	@ If the user ever returns, go to an infinte loop
	ldr	r0, =ILoop
	ldr	r0, [r0]
	ldr	r1, =0x027FFE78
	str	r0, [r1]
	bx	r1
ILoop:
	b	ILoop

@---------------------------------------------------------------------------------
@ Clear memory to 0x00 if length != 0
@  r0 = Start Address
@  r1 = Length
@---------------------------------------------------------------------------------
ClearMem:
@---------------------------------------------------------------------------------
	mov	r2, #3			@ Round down to nearest word boundary
	add	r1, r1, r2		@ Shouldn't be needed
	bics	r1, r1, r2		@ Clear 2 LSB (and set Z)
	bxeq	lr			@ Quit if copy size is 0

	mov	r2, #0
ClrLoop:
	stmia	r0!, {r2}
	subs	r1, r1, #4
	bne	ClrLoop

	bx	lr

@---------------------------------------------------------------------------------
@ Copy memory if length	!= 0
@  r1 = Source Address
@  r2 = Dest Address
@  r4 = Dest Address + Length
@---------------------------------------------------------------------------------
CopyMemCheck:
@---------------------------------------------------------------------------------
	sub	r3, r4, r2		@ Is there any data to copy?
@---------------------------------------------------------------------------------
@ Copy memory
@  r1 = Source Address
@  r2 = Dest Address
@  r3 = Length
@---------------------------------------------------------------------------------
CopyMem:
@---------------------------------------------------------------------------------
	mov	r0, #3			@ These commands are used in cases where
	add	r3, r3, r0		@ the length is not a multiple of 4,
	bics	r3, r3, r0		@ even though it should be.
	bxeq	lr			@ Length is zero, so exit
CIDLoop:
	ldmia	r1!, {r0}
	stmia	r2!, {r0}
	subs	r3, r3, #4
	bne	CIDLoop

	bx	lr
	
initMem:
	adr r12,memconfig
	
	ldr r0,=0x00002078
	mcr p15,0,r0,c1,c0,0	@cache+TCM+PU off
	
	ldmia r12!,{r0-r7}
	mcr p15,0,r0,c6,c0		@regions
	mcr p15,0,r1,c6,c1
	mcr p15,0,r2,c6,c2
	mcr p15,0,r3,c6,c3
	mcr p15,0,r4,c6,c4
	mcr p15,0,r5,c6,c5
	mcr p15,0,r6,c6,c6
	mcr p15,0,r7,c6,c7
	
	ldmia r12!,{r0-r6}
	mcr p15,0,r0,c2,c0,0	@cache data
	mcr p15,0,r1,c3,c0,0	@buffer data
	mcr p15,0,r2,c2,c0,1	@cache inst
	mcr p15,0,r3,c5,c0,2	@data access
	mcr p15,0,r4,c5,c0,3	@inst access
	mcr p15,0,r5,c9,c1,0	@dtcm base
	mcr p15,0,r6,c9,c1,1	@itcm base
	
	;ldr r0,=0x52078
	ldr r0,=0x5707d
	mcr p15,0,r0,c1,c0,0	@cache+TCM+PU on
	mov r0,#0
	mcr p15,0,r0,c7,c5,0	@icache flush
	mcr p15,0,r0,c7,c6,0	@dcache flush
	mov pc,lr
	
memconfig:
	.word 0x01ff801d	@->1ff8000 (32K) ITCM
	.word 0x0200002b	@->2000000 (4M) main
	.word 0x04000033	@->4000000 (64M) regs
	.word 0x08000031	@->8000000 (64M) cart
	.word 0xffff0017	@->ffff1000 (4K) bios
	.word 0x0240002b	@->2400000 (4M) main (uncached)
	.word 0x0280001b	@(16K) DTCM
	.word 0x0a00001f	@(64k) SRAM

	.word (0x8|0x2)		@data cache
	.word 0x2			@data buffer
	.word 0x2			@inst cache
	.word 0x33363333	@data access (0=NA, 3=RW, 6=R)
	.word 0x00063033	@inst access
	.word 0x0280000a	@(16K) DTCM base
	.word 0x00000020	@->2000000 (32M) ITCM base

@---------------------------------------------------------------------------------
	.align
	.pool
	.end
@---------------------------------------------------------------------------------
