#include "../../debug.h"

#ifdef COMPILE_SNEESE

#include <NDS/NDS.h>
#include "spc.h"
#include "../sprintf.h"
#include "../debug.h"

void memcpy(void *dest, const void *src, int length);

unsigned char SPCRAM[0x10000];
unsigned SPC_DSP_DATA;
unsigned char SPC_DSP[256];

void Wrap_SDSP_Cyclecounter() {
}

void SPC_READ_DSP() {
}
void SPC_WRITE_DSP() {
}

void SPC_SetState(int pc,int a,int x,int y,int p,int sp,void *ram);
void initSneeseApu(const u8 *spc) {
	SPC_SetState(spc[0x25] | (spc[0x26] << 8),spc[0x27],spc[0x28],spc[0x29],
	  spc[0x2A],0x100+spc[0x2B],(void*)&spc[0x100]);
	memcpy(SPC_DSP,&spc[0x10100],128);
}

void SneeseDebugGetSPCState(char *buffer) {
    u8 p = get_SPC_PSW();
    sprintf (buffer, "A:%02X X:%02X Y:%02X S:%3X DSP:%02X\nP:%c%c%c%c%c%c%c%c T0:%02X T1:%02X T2:%02X\nCont: %02X C0: %02X C1: %02X C2: %02X",
		_SPC_A, _SPC_X, _SPC_Y,
		0x100 | _SPC_SP,
        SPCRAM[0xF2],
		p & 0x80 ? 'N' : 'n',
		p & 0x40 ? 'V' : 'v',
		p & 0x20 ? 'P' : 'p',
		0 ? 'B' : 'b', // Break flag
		p & 0x08 ? 'H' : 'h', // Half carry flag
		0 ? 'I' : 'i', // Interrupt flag
		p & 0x02 ? 'Z' : 'z',
		p & 0x01 ? 'C' : 'c',
		SPCRAM[0xFA],
		SPCRAM[0xFB],
		SPCRAM[0xFC],
		SPCRAM[0xF1],
		SPCRAM[0xFD],
		SPCRAM[0xFE],
		SPCRAM[0xFF]);
}

extern u8 *APU_MEM;
extern u32 APU_STATE[16];
extern u16 APU_SP;
u8 MakeRawPSWFromState(u32 state[16]);

bool compareState(u8 *mem, u8 *dsp, u8 *extra, u16 pc, char *debugState) {
    bool good = true;
    char *messup = "";
    char buf[100];

//    if (last9xCycles != SPC_IPC->lastApuCycles) { good = false; sprintf(buf, "Cycles %d, %d", SPC_IPC->lastApuCycles, last9xCycles); messup = buf; }
    if (_SPC_A != APU_STATE[0] >> 24) { good = false; messup = "A"; }
    if (_SPC_X != APU_STATE[1] >> 24) { good = false; messup = "X"; }
    if (_SPC_Y != APU_STATE[2] >> 24) { good = false; messup = "Y"; }
    if ((0x100 | _SPC_SP) != APU_SP) { good = false; sprintf(buf, "SP %x %x", _SPC_SP, APU_SP); messup = buf; }
    u8 p = get_SPC_PSW() & ~0x10;
    u8 myp = MakeRawPSWFromState(APU_STATE);
    if (p != myp) { good = false; sprintf(buf, "Flags P:%x P:%x", p, myp); messup = buf; }
    if (_SPC_PC != APU_STATE[5] - APU_STATE[3]) { good = false; sprintf(buf, "PC %x, %x", APU_STATE[5] - APU_STATE[3], _SPC_PC); messup = buf; }

    if (!good) {
        char buf[150], tbuf[150], tbuf2[50];
        SneeseDebugGetSPCState(buf);
        DebugGetSPCState(tbuf);
        DebugGetOpcodeDesc(tbuf2, &(mem[pc]), true, mem);
        sprintf(debugState, "Cpu State differs: %s\n%s\n%s\n%s",messup,tbuf2,tbuf,buf);
        return false;
    }

    u32 *ptr1 = (u32*)mem;
    u32 *ptr2 = (u32*)SPCRAM;
    for (int i = 0; i < 0x10000 / 4; i++) {
        if (ptr1[i] != ptr2[i]) {
            for (int j = 0; j < 4; j++) {
                if (mem[(i*4)+j] != SPCRAM[(i*4)+j]) {
                    char buf[150], tbuf[150], tbuf2[50];
                    SneeseDebugGetSPCState(buf);
                    DebugGetSPCState(tbuf);
                    DebugGetOpcodeDesc(tbuf2, &(SPCRAM[pc]), true, SPCRAM);
                    sprintf(debugState, "Mem differs at %x: %x-%x\n%s\n%s\n%s",i*4+j,mem[i*4+j],SPCRAM[i*4+j],tbuf2,tbuf,buf);
                    return false;
                }
            }
        }
    }
/*
    for (int i=0; i<=0x7f; i++) {
        if (dsp[i] != APU.DSP[i]) {
            char buf[150], tbuf[150], tbuf2[50];
            S9xDebugGetSPCState(buf);
            DebugGetSPCState(tbuf);
            DebugGetOpcodeDesc(tbuf2, &(IAPU.RAM[pc]), true, IAPU.RAM);
            sprintf(debugState, "DSP differs at %x: %x-%x\n%s\n%s\n%s",i,dsp[i],APU.DSP[i],tbuf2,tbuf,buf);
            return false;
        }
    }

    for (int i=0; i<=0x3f; i++) {
        if (APU.ExtraRAM[i] != extra[i]) {
            char buf[150], tbuf[150], tbuf2[50];
            S9xDebugGetSPCState(buf);
            DebugGetSPCState(tbuf);
            DebugGetOpcodeDesc(tbuf2, &(IAPU.RAM[pc]), true, IAPU.RAM);
            sprintf(debugState, "Extra mem differs at %x: %x-%x\n%s\n%s\n%s",i,extra[i],APU.ExtraRAM[i],tbuf2,tbuf,buf);
            return false;
        }
    }*/

//    for (int i = 0; i < 8; i++) {
//        if (s9xChannels[0].samplePos != ((DspChannel*)SPC_IPC->channelDebug)->samplePos)

    return true;
}
#endif
