#include "../../spcme.h"

#ifdef COMPILE_SNES9X

#include "snes9x.h"
#include "spc700.h"
#include "apu.h"
#include "dsp.h"

// Envelope timing table.  Number of counts that should be subtracted from the counter
// at each sample period.  From libopenspc source -- should check if this is correct.
// The counter starts at 30720 (0x7800).
static const s16 ENVCNT_START = 0x7800;
static const s16 ENVCNT[0x20] = {
	0x0000,0x000F,0x0014,0x0018,0x001E,0x0028,0x0030,0x003C,
	0x0050,0x0060,0x0078,0x00A0,0x00C0,0x00F0,0x0140,0x0180,
	0x01E0,0x0280,0x0300,0x03C0,0x0500,0x0600,0x0780,0x0A00,
	0x0C00,0x0F00,0x1400,0x1800,0x1E00,0x2800,0x3C00,0x7800
};

u8 s9xDSP_MEM[0x80];
s9xDspChannel s9xChannels[8];

#ifdef ACCURATE_MIXER
void s9xDecodeSampleBlock(s9xDspChannel *channel) {
    if (channel->blockPos > 0x10000 - 9) {
        channel->lastBlock = true;
        channel->loop = false;
        return;
    }

    s8 *cur = (s8*)&(IAPU.RAM[channel->blockPos]);
    s16 *sample = channel->decoded;

    if (*cur & 1) {
        // Last block
        channel->lastBlock = true;
        channel->loop = *cur & 2 != 0;
    }

    s16 prev1 = channel->prevSamp1, prev2 = channel->prevSamp2;
    
    {
        s8 shift = 28 - ((*cur++ >> 4) & 0xF);

		if (shift == 28) { // if the range was 0
			for (int i=8; i!=0; i--) {
				*sample++ = (s16)(((s32)*cur >> 4) & -2);
				*sample++ = (s16)((((s32)*cur << 28) >> 28) & -2);
				cur++;
			}
		} else if (shift >= 16) { // if the range was 1 through 12
			for (int i=8; i!=0; i--) {
				*sample++ = (s16)((((s32)*cur & 0xF0) << 24) >> shift);
				*sample++ = (s16)(((s32)*cur << 28) >> shift);
				cur++;
			}
		} else { // invalid range
			for (int i=8; i!=0; i--) {
				*sample++ = (*cur & 0x80) ? -4096 : 0;
				*sample++ = (*cur & 8) ? -4096 : 0;
				cur++;
			}
		}

		sample -= 16;

		switch (cur[-9] & 0xC) {
		case 0x0:
			prev1 = sample[15];
			prev2 = sample[14];
            sample += 16;
			break;

		case 0x4:
			for (int i=16; i!=0; i--) {
				*sample += (s16)(prev1 + ((-prev1 >> 4) & -2));
				prev1 = *sample++;
			}
			prev2 = sample[-2];
			break;

		case 0x8: //Filter 2:  s = d + prev1(61/32) - prev2(15/16) 
			for (int i=16; i!=0; i--) 
			{ 
				s32 s = ((prev1 << 1) + (((-prev1 * 3) >> 5) & -2)) + 
					(-prev2 + ((prev2 >> 4) & -2)) + 
					*sample; 

				if (s < -65536) s = -65536;     //Clamp result to 17 bits 
				else if (s > 65534) s = 65534; 

				prev2 = prev1; 
				prev1 = *sample++ = (s16)s; 
			} 
			break; 

		case 0xC: //Filter 3:  s = d + prev1(115/64) - prev2(13/16) 
			for (int i=16; i!=0; i--) 
			{ 
				s32 s = ((prev1 << 1) + (((-13 * prev1) >> 6) & -2)) + 
					(-prev2 + (((prev2 * 3) >> 4) & -2)) + 
					*sample; 

				if (s < -65536) s = -65536; 
				else if (s > 65534) s = 65534; 

				prev2 = prev1; 
				prev1 = *sample++ = (s16)s; 
			} 
			break; 
		}
	}

    channel->prevSamp1 = prev1;
    channel->prevSamp2 = prev2;

    channel->blockPos += 9;
}
#else
void s9xDecodeSampleBlock(s9xDspChannel *channel) {
    if (channel->blockPos > 0x10000 - 9) {
        channel->lastBlock = true;
        channel->loop = false;
        return;
    }

    s8 *cur = (s8*)&(IAPU.RAM[channel->blockPos]);
    s16 *sample = channel->decoded;

    if (*cur & 1) {
        // Last block
        channel->lastBlock = true;
        channel->loop = *cur & 2 != 0;
    }

/*#ifndef SLOW_MIXER
    u32 prevSamples = DecodeSampleBlockAsm(cur, sample, channel->prevSamp1, channel->prevSamp2);

    channel->prevSamp1 = prevSamples & 0xffff;
    channel->prevSamp2 = prevSamples >> 16;
#else*/
    s16 prev0 = channel->prevSamp1, prev1 = channel->prevSamp2;

    u8 filter = *cur++;
    u8 shift = filter >> 4;
	
    switch ((filter >> 2) & 3)
    {
    case 0:
		for (s8 i = 8; i != 0; i--)
		{
			s8 sample1 = *cur++;
			s8 sample2 = sample1 << 4;
			sample2 >>= 4;
			sample1 >>= 4;
			*sample++ = ((int32) sample1 << shift);
			*sample++ = ((int32) sample2 << shift);
		}
		prev1 = *(sample - 2);
		prev0 = *(sample - 1);
		break;
    case 1:
		for (s8 i = 8; i != 0; i--)
		{
			s8 sample1 = *cur++;
			s8 sample2 = sample1 << 4;
			sample2 >>= 4;
			sample1 >>= 4;
			prev0 = (int16) prev0;
			*sample++ = prev1 = ((int32) sample1 << shift) + prev0 - (prev0 >> 4);
			prev1 = (int16) prev1;
			*sample++ = prev0 = ((int32) sample2 << shift) + prev1 - (prev1 >> 4);
		}
		break;
    case 2:
		for (s8 i = 8; i != 0; i--)
		{
			s8 sample1 = *cur++;
			s8 sample2 = sample1 << 4;
			sample2 >>= 4;
			sample1 >>= 4;
			
			s32 out = (sample1 << shift) - prev1 + (prev1 >> 4);
			prev1 = (int16) prev0;
			prev0 &= ~3;
			*sample++ = prev0 = out + (prev0 << 1) - (prev0 >> 5) - 
				(prev0 >> 4);
			
			out = (sample2 << shift) - prev1 + (prev1 >> 4);
			prev1 = (int16) prev0;
			prev0 &= ~3;
			*sample++ = prev0 = out + (prev0 << 1) - (prev0 >> 5) -
				(prev0 >> 4);
		}
		break;
    case 3:
		for (s8 i = 8; i != 0; i--)
		{
			s8 sample1 = *cur++;
			s8 sample2 = sample1 << 4;
			sample2 >>= 4;
			sample1 >>= 4;
			s32 out = (sample1 << shift);

			out = out - prev1 + (prev1 >> 3) + (prev1 >> 4);
			prev1 = (int16) prev0;
			prev0 &= ~3;
			*sample++ = prev0 = out + (prev0 << 1) - (prev0 >> 3) - 
				(prev0 >> 4) - (prev1 >> 6);

			out = (sample2 << shift);
			out = out - prev1 + (prev1 >> 3) + (prev1 >> 4);
			prev1 = (int16) prev0;
			prev0 &= ~3;
			*sample++ = prev0 = out + (prev0 << 1) - (prev0 >> 3) - 
				(prev0 >> 4) - (prev1 >> 6);
		}
		break;
    }

    channel->prevSamp1 = prev0;
    channel->prevSamp2 = prev1;

    channel->blockPos += 9;
}
#endif

void s9xDspReset() {
	for (int i = 0; i < 8; i++) {
        s9xChannels[i].samplePos = 0;
        s9xChannels[i].envCount = 0;
        s9xChannels[i].needsDecode = true;
        s9xChannels[i].lastBlock = false;
        s9xChannels[i].active = false;
	}
}

void s9xDspSetChannelVolume(u32 channel) {
    s9xChannels[channel].leftVolume = s9xDSP_MEM[(channel << 4) + DSP_VOL_L];
    s9xChannels[channel].rightVolume = s9xDSP_MEM[(channel << 4) + DSP_VOL_R];

    if (s9xChannels[channel].leftVolume < 0) s9xChannels[channel].leftVolume = -s9xChannels[channel].leftVolume;
    if (s9xChannels[channel].rightVolume < 0) s9xChannels[channel].rightVolume = -s9xChannels[channel].rightVolume;
}

void s9xDspSetChannelPitch(u32 channel) {
	u16 rawFreq = ((s9xDSP_MEM[(channel << 4) + DSP_PITCH_H] << 8) + s9xDSP_MEM[(channel << 4) + DSP_PITCH_L]) & 0x3fff;

	s9xChannels[channel].sampleSpeed = ((rawFreq << 3) << 12) / MIXRATE;
}

void s9xDspSetChannelSource(u32 channel) {
	u8 *sampleBase = IAPU.RAM + (s9xDSP_MEM[DSP_DIR] << 8);
	u16 *lookupPointer = (u16*)(sampleBase + (s9xDSP_MEM[(channel << 4) + DSP_SRC] << 2));

    s9xChannels[channel].blockPos = lookupPointer[0];
    s9xChannels[channel].blockLoop = lookupPointer[1];
}

void s9xDspSetEndOfSample(u32 channel) {
    s9xChannels[channel].active = false;

    s9xDSP_MEM[DSP_KON] &= ~(1 << channel);
    s9xDSP_MEM[DSP_KOF] &= ~(1 << channel);
    s9xDSP_MEM[DSP_ENDX] |= 1 << channel;
    s9xDSP_MEM[(channel << 4) | DSP_ENVX] = 0;
    s9xDSP_MEM[(channel << 4) | DSP_OUTX] = 0;
}

void s9xDspSetChannelEnvelopeHeight(u32 channel, u8 height, bool offCheck) {
    s9xChannels[channel].envx = (u8)height;

    s9xChannels[channel].leftCalcVolume = (s9xChannels[channel].leftVolume * height) / 128;
    s9xChannels[channel].rightCalcVolume = (s9xChannels[channel].rightVolume * height) / 128;

    if (offCheck && s9xChannels[channel].envx == 0 && s9xChannels[channel].envState != ENVSTATE_DIRECT)
    {
		s9xDspSetEndOfSample(channel);
    }
}

void s9xDspSetChannelEnvelope(u32 channel) {
	u8 adsr1 = s9xDSP_MEM[(channel << 4) + DSP_ADSR1];
	u8 adsr2 = s9xDSP_MEM[(channel << 4) + DSP_ADSR2];
	u8 gain = s9xDSP_MEM[(channel << 4) + DSP_GAIN];

	if (adsr1 & 0x80) {
		// ADSR mode, set envelope up
		// Attack rate goes into envelopeSpeed initially
		u8 attack = adsr1 & 0xf;
		u8 decay = (adsr1 >> 4) & 0x7;
		u8 sustainLevel = adsr2 >> 5;
		u8 sustainRate = adsr2 & 0x1f;

		s9xChannels[channel].decaySpeed = ENVCNT[(decay << 1) + 0x10];
		s9xChannels[channel].sustainLevel = 0x10 * (sustainLevel + 1);
		s9xChannels[channel].sustainSpeed = ENVCNT[sustainRate];

		if (attack == 0xf) {
			// 0ms attack, go straight to full volume, and set decay
			s9xDspSetChannelEnvelopeHeight(channel, 0x7f, false);
			s9xChannels[channel].envSpeed = s9xChannels[channel].decaySpeed;
			s9xChannels[channel].envState = ENVSTATE_DECAY;
		} else {
			s9xDspSetChannelEnvelopeHeight(channel, 0, false);
			s9xChannels[channel].envSpeed = ENVCNT[(attack << 1) + 1];
			s9xChannels[channel].envState = ENVSTATE_ATTACK;
		}
	} else {
		// Gain mode
        if (s9xChannels[channel].envState != ENVSTATE_RELEASE) {
            if ((gain & 0x80) == 0) {
                // Direct designation
                s9xChannels[channel].envState = ENVSTATE_DIRECT;
                s9xDspSetChannelEnvelopeHeight(channel, gain & 0x7f, true);
            } else {
                s9xChannels[channel].envSpeed = ENVCNT[gain & 0x1f];

                // In this case, envelope volume remains at what it was when we entered the state
                switch ((gain >> 5) & 0x3) {
                case 0:
                    // Linear decrease
                    s9xChannels[channel].envState = ENVSTATE_DECREASE;
                    break;
                case 1:
                    // Exponential decrease
                    s9xChannels[channel].envState = ENVSTATE_DECEXP;
                    break;
                case 2:
                    // Linear increase
                    s9xChannels[channel].envState = ENVSTATE_INCREASE;
                    break;
                case 3:
                    // Bent line increase
                    s9xChannels[channel].envState = ENVSTATE_BENTLINE;
                    break;
                }
            }
		}
	}
}

void s9xDspWriteByte(u8 val, u8 address) {
    if (address > 0x7f) return;

	bool sameWritten = s9xDSP_MEM[address] == val;

    s9xDSP_MEM[address] = val;

	switch (address) {
		case 0x00 + DSP_VOL_L:
		case 0x10 + DSP_VOL_L:
		case 0x20 + DSP_VOL_L:
		case 0x30 + DSP_VOL_L:
		case 0x40 + DSP_VOL_L:
		case 0x50 + DSP_VOL_L:
		case 0x60 + DSP_VOL_L:
		case 0x70 + DSP_VOL_L:
			s9xDspSetChannelVolume(address >> 4);
			break;

		case 0x00 + DSP_VOL_R:
		case 0x10 + DSP_VOL_R:
		case 0x20 + DSP_VOL_R:
		case 0x30 + DSP_VOL_R:
		case 0x40 + DSP_VOL_R:
		case 0x50 + DSP_VOL_R:
		case 0x60 + DSP_VOL_R:
		case 0x70 + DSP_VOL_R:
			s9xDspSetChannelVolume(address >> 4);
			break;

		case 0x00 + DSP_PITCH_L:
		case 0x10 + DSP_PITCH_L:
		case 0x20 + DSP_PITCH_L:
		case 0x30 + DSP_PITCH_L:
		case 0x40 + DSP_PITCH_L:
		case 0x50 + DSP_PITCH_L:
		case 0x60 + DSP_PITCH_L:
		case 0x70 + DSP_PITCH_L:
			s9xDspSetChannelPitch(address >> 4);
			break;

		case 0x00 + DSP_PITCH_H:
		case 0x10 + DSP_PITCH_H:
		case 0x20 + DSP_PITCH_H:
		case 0x30 + DSP_PITCH_H:
		case 0x40 + DSP_PITCH_H:
		case 0x50 + DSP_PITCH_H:
		case 0x60 + DSP_PITCH_H:
		case 0x70 + DSP_PITCH_H:
			s9xDspSetChannelPitch(address >> 4);
			break;

		case 0x00 + DSP_SRC:
		case 0x10 + DSP_SRC:
		case 0x20 + DSP_SRC:
		case 0x30 + DSP_SRC:
		case 0x40 + DSP_SRC:
		case 0x50 + DSP_SRC:
		case 0x60 + DSP_SRC:
		case 0x70 + DSP_SRC:
            if (!sameWritten)
    			s9xDspSetChannelSource(address >> 4);
			break;

		case 0x00 + DSP_ADSR1:
		case 0x10 + DSP_ADSR1:
		case 0x20 + DSP_ADSR1:
		case 0x30 + DSP_ADSR1:
		case 0x40 + DSP_ADSR1:
		case 0x50 + DSP_ADSR1:
		case 0x60 + DSP_ADSR1:
		case 0x70 + DSP_ADSR1:
			if (!sameWritten)
				s9xDspSetChannelEnvelope(address >> 4);
			break;

		case 0x00 + DSP_ADSR2:
		case 0x10 + DSP_ADSR2:
		case 0x20 + DSP_ADSR2:
		case 0x30 + DSP_ADSR2:
		case 0x40 + DSP_ADSR2:
		case 0x50 + DSP_ADSR2:
		case 0x60 + DSP_ADSR2:
		case 0x70 + DSP_ADSR2:
			if (!sameWritten)
				s9xDspSetChannelEnvelope(address >> 4);
			break;

		case 0x00 + DSP_GAIN:
		case 0x10 + DSP_GAIN:
		case 0x20 + DSP_GAIN:
		case 0x30 + DSP_GAIN:
		case 0x40 + DSP_GAIN:
		case 0x50 + DSP_GAIN:
		case 0x60 + DSP_GAIN:
		case 0x70 + DSP_GAIN:
			if (!sameWritten)
				s9xDspSetChannelEnvelope(address >> 4);
			break;

		case DSP_KON:
//            val &= ~s9xDSP_MEM[DSP_KOF];
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
			        s9xDspSetChannelVolume(i);
        			s9xDspSetChannelPitch(i);
                    s9xDspSetChannelSource(i);
                    s9xDspSetChannelEnvelope(i);

                    s9xChannels[i].samplePos = 0;
                    s9xChannels[i].envCount = 0;

                    s9xChannels[i].needsDecode = true;
					s9xChannels[i].envCount = ENVCNT_START;
					s9xChannels[i].active = true;

                    if ((s9xDSP_MEM[DSP_NOV]>>i)&1) {
    					s9xChannels[i].active = false;
                        // Noise sample
                    } else {
                        // Normal sample
                    }

                    s9xDSP_MEM[DSP_ENDX] &= ~(1 << i);
				}
			break;

		case DSP_KOF:
			for (int i=0; i<8; i++)
				if (((val>>i)&1) && s9xChannels[i].active) {
					// Set current state to release (ENDX will be set when release hits 0)
					s9xChannels[i].envState = ENVSTATE_RELEASE;
					s9xChannels[i].envCount = ENVCNT_START;
                    s9xChannels[i].envSpeed = ENVCNT[0x1C];

					// Clear KON register.
					s9xDSP_MEM[DSP_KON] &= ~(1 << i);
				}
			break;

		case DSP_PMOD:
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
//					dspunimpl(DSP_PMOD);
//					s9xChannels[i].active = 0;
				}
			break;

		case DSP_NOV:
			for (int i=0; i<8; i++)
				if ((val>>i)&1) {
//					dspunimpl(DSP_NOV);
					s9xChannels[i].active = false;
				}
			break;

		case DSP_ENDX:
			s9xDSP_MEM[DSP_ENDX] = 0;
			break;

/*		case DSP_DIR:
            if (!sameWritten)
    			for (int i = 0; i < 8; i++) {
	    			s9xDspSetChannelSource(i);
		    	}
			break;
*/

		// Echo filter not supported
		case 0x0F:
		case 0x1F:
		case 0x2F:
		case 0x3F:
		case 0x4F:
		case 0x5F:
		case 0x6F:
		case 0x7F:
			break;

		default:
			break;
	}
}

void s9xDspPrepareStateAfterReload() {
	for (u32 i = 0; i < 8; i++) {
		if (s9xDSP_MEM[DSP_KON] & (1 << i)) {
            // Set up the state here
            s9xDspSetChannelVolume(i);
            s9xDspSetChannelPitch(i);
            s9xDspSetChannelSource(i);
            s9xDspSetChannelEnvelope(i);

            s9xChannels[i].needsDecode = true;
            s9xChannels[i].envCount = ENVCNT_START;
            s9xChannels[i].active = true;

            if ((s9xDSP_MEM[DSP_NOV]>>i)&1) {
                // Noise sample
                s9xChannels[i].active = false;
            } else {
                // Normal sample
            }
        }
	}
}

s32 mixLeft[MIXBUFSIZE] EWRAM_BSS;
s32 mixRight[MIXBUFSIZE] EWRAM_BSS;

void s9xDspMixSamplesStereo(int numSamples, u16 *left, u16 *right) {
    for (int i = 0; i < numSamples; i++) {
        mixLeft[i] = 0;
        mixRight[i] = 0;
    }

    for (int chan = 0; chan < 8; chan++) {
        s9xDspChannel *channel = &s9xChannels[chan];

        if (channel->active == false) goto nextChannel;

        if (channel->needsDecode) {
            channel->needsDecode = false;

            s9xDecodeSampleBlock(channel);
        }

        for (int i = 0; i < numSamples; i++) {
            channel->envCount -= channel->envSpeed;
            if (channel->envCount < 0) {
                channel->envCount = ENVCNT_START;

                switch (channel->envState)
                {
                case ENVSTATE_ATTACK:
                    channel->envx += 4;
                    if (channel->envx >= 0x7f) {
                        channel->envx = 0x7f;
                        channel->envState = ENVSTATE_DECAY;
                        channel->envSpeed = channel->decaySpeed;
                    }
                    break;

                case ENVSTATE_DECAY:
                    channel->envx = (channel->envx * 255) >> 8;
                    if (channel->envx <= channel->sustainLevel) {
                        if (channel->envx <= 0) {
                            s9xDspSetEndOfSample(chan);
                            goto nextChannel;
                        }

                        channel->envState = ENVSTATE_SUSTAIN;
                        channel->envSpeed = channel->sustainSpeed;
                    }
                    break;

                case ENVSTATE_SUSTAIN:
                    channel->envx = (channel->envx * 255) >> 8;

                    if (channel->envx <= 0) {
                        s9xDspSetEndOfSample(chan);
                        goto nextChannel;
                    }
                    break;

                case ENVSTATE_RELEASE:
                    channel->envx--;

                    if (channel->envx <= 0) {
                        s9xDspSetEndOfSample(chan);
                        goto nextChannel;
                    }
                    break;

                case ENVSTATE_INCREASE:
                    channel->envx += 4;

                    if (channel->envx > 0x7f) {
                        channel->envx = 0x7f;
                        channel->envState = ENVSTATE_DIRECT;
                        channel->envCount = ENVCNT_START;
                        channel->envSpeed = 0;
                    }
                    break;

                case ENVSTATE_BENTLINE:
                    if (channel->envx > 0x5f) {
                        channel->envx++;
                    } else {
                        channel->envx += 4;
                    }

                    if (channel->envx > 0x7f) {
                        channel->envx = 0x7f;
                        channel->envState = ENVSTATE_DIRECT;
                        channel->envCount = ENVCNT_START;
                        channel->envSpeed = 0;
                    }
                    break;

                case ENVSTATE_DECREASE:
                    channel->envx -= 4;

                    if (channel->envx <= 0) {
                        s9xDspSetEndOfSample(chan);
                        goto nextChannel;
                    }
                    break;

                case ENVSTATE_DECEXP:
                    channel->envx = (channel->envx * 255) >> 8;

                    if (channel->envx <= 0) {
                        s9xDspSetEndOfSample(chan);
                        goto nextChannel;
                    }
                    break;
                }

                channel->leftCalcVolume = (channel->leftVolume * channel->envx) / 128;
                channel->rightCalcVolume = (channel->rightVolume * channel->envx) / 128;
            }

            channel->samplePos += channel->sampleSpeed;
            if (channel->samplePos >= (16 << 12)) {
                channel->samplePos -= 16 << 12;

                if (channel->lastBlock) {
                    if (!channel->loop) {
                        // End the channel
                        s9xDspSetEndOfSample(chan);
                        goto nextChannel;
                    } else {
                        channel->lastBlock = false;
                        channel->blockPos = channel->blockLoop;
                        s9xDSP_MEM[DSP_ENDX] |= (1 << chan);
                    }
                }
                s9xDecodeSampleBlock(channel);
            }

            // Pitch modulation should be done here

            mixLeft[i] += (channel->decoded[channel->samplePos] * channel->leftCalcVolume) >> 8;
            mixRight[i] += (channel->decoded[channel->samplePos] * channel->rightCalcVolume) >> 8;

            // Echo buffer needs to be done here as well...
        }

nextChannel:

        // Set ENVX and OUTX
        s9xDSP_MEM[(chan << 4) + DSP_ENVX] = channel->envx;
        s9xDSP_MEM[(chan << 4) + DSP_OUTX] = (channel->decoded[channel->samplePos >> 12] * channel->envx) >> (7 + 8);
    }

    for (int i = 0; i < numSamples; i++) {
        if (mixLeft[i] < -32768) mixLeft[i] = -32768;
        if (mixLeft[i] > 32767) mixLeft[i] = 32767;
        if (mixRight[i] < -32768) mixRight[i] = -32768;
        if (mixRight[i] > 32767) mixRight[i] = 32767;

        left[i] = mixLeft[i];
        right[i] = mixRight[i];
    }
}

#endif
