#include "../../../spcme.h"
#ifdef COMPILE_SNES9X

#include "snes9x.h"
#include "apu.h"
#include "spc700.h"
#include "dsp.h"

#include "../../../arm7/source/apu.h"
#include "../debug.h"
#include "../sprintf.h"
#include <NDS/ARM9/console.h> //basic print funcionality

void WaitForVBlank();
u8 s9xRam[0x10000];
u8 s9xApuSleeping;

void memset(void *data, int fill, int length);
void memcpy(void *dest, const void *src, int length);

void S9xDebugGetSPCState(char *buffer) {
    sprintf (buffer, "A:%02X X:%02X Y:%02X S:%3X DSP:%02X\nP:%c%c%c%c%c%c%c%c T0:%02X T1:%02X T2:%02X\nCont: %02X C0: %02X C1: %02X C2: %02X",
		APURegisters.YA.B.A, APURegisters.X, APURegisters.YA.B.Y,
		0x100 | APURegisters.S,
        IAPU.RAM[0xF2],
		IAPU._Zero & 0x80 ? 'N' : 'n',
		IAPU._Overflow ? 'V' : 'v',
		APURegisters.P & 0x20 ? 'P' : 'p',
		0 ? 'B' : 'b', // Break flag
		((APURegisters.P >> 3) & 1) ? 'H' : 'h', // Half carry flag
		0 ? 'I' : 'i', // Interrupt flag
		IAPU._Zero == 0 ? 'Z' : 'z',
		IAPU._Carry ? 'C' : 'c',
		IAPU.RAM[APU_TIMER0],
		IAPU.RAM[APU_TIMER1],
		IAPU.RAM[APU_TIMER2],
	    IAPU.RAM[APU_CONTROL_REG],
		IAPU.RAM[APU_COUNTER0],
		IAPU.RAM[APU_COUNTER1],
		IAPU.RAM[APU_COUNTER2]);
}

void S9xPrintDisasm(int num, u16 pc) {
    char buf[150];
    S9xDebugGetSPCState(buf);

    consolePrintSet(0,0);
    for (int i = 0; i < num + 3; i++) {
        consolePrintf("                                              \n");
    }

    consolePrintSet(0,0);
    consolePrintf("%s\n", buf);

    while (num--) {
        pc += DebugGetOpcodeDesc(buf, &(IAPU.RAM[pc]), true, IAPU.RAM);
        consolePrintf("%s\n", buf);
    }
}

void init9xApu(const u8 *spc) {
    s9xApuSleeping = 0;

    IAPU.RAM = s9xRam;
    APU.DSP = s9xDSP_MEM;

    s9xDspReset();

    memcpy(IAPU.RAM, &spc[0x100], 0x10000);
    IAPU.PC = IAPU.RAM + (spc[0x25] | (spc[0x26] << 8));
    IAPU.DirectPage = IAPU.RAM;
    if (spc[0x2A] & 0x20) IAPU.DirectPage += 0x100;
 
    IAPU.APUExecuting = 1;
    APU.ShowROM = TRUE;

    IAPU.OneCycle = 1;
    IAPU.TwoCycles = 2;

    APU.Cycles = 0;
    APURegisters.YA.B.A = spc[0x27];
    APURegisters.YA.B.Y = spc[0x29];
    APURegisters.X = spc[0x28];
    APURegisters.S = spc[0x2B];
    APURegisters.P = spc[0x2A];
    S9xAPUUnpackStatus ();
    APURegisters.PC = spc[0x25] | (spc[0x26] << 8);

    for (int i=0; i<=0x7f; i++) {
        APU.DSP[i] = spc[0x10100 + i];
    }

    for (int i=0; i<=0x3f; i++) APU.ExtraRAM[i] = spc[0x101c0 + i];

    APU.TimerEnabled [0] = IAPU.RAM[APU_CONTROL_REG] & 1;
    APU.TimerEnabled [1] = (IAPU.RAM[APU_CONTROL_REG] & 2) >> 1;
    APU.TimerEnabled [2] = (IAPU.RAM[APU_CONTROL_REG] & 4) >> 2;

    IAPU.RAM[APU_COUNTER0] &= 0xf;
    IAPU.RAM[APU_COUNTER1] &= 0xf;
    IAPU.RAM[APU_COUNTER2] &= 0xf;

    for (int i = 0; i < 3; i++) {
        APU.Timer[i] = 0;   
        APU.TimerCycles[i] = 0;
        APU.TimerTarget [i] = IAPU.RAM [0xfa + i];
        if (APU.TimerTarget [i] == 0) APU.TimerTarget [i] = 0x100;
    }

    for (int i = 0; i < 4; i++) {
        APU.InPorts[i] = IAPU.RAM[0xf4 + i];
    }

    APU.ShowROM = IAPU.RAM[APU_CONTROL_REG] >> 7;
    S9xSetAPUControl(IAPU.RAM[APU_CONTROL_REG]);
    s9xDspPrepareStateAfterReload();
}

void updateTimer9x(int timer, int cyclesRun) {
    int shift;
    if (timer == 2) {
        shift = t64Shift;
    } else {
        shift = t8Shift;
    }
    int mask = (1 << shift) - 1;

    APU.TimerCycles[timer] += cyclesRun;
    int updates = APU.TimerCycles[timer] >> shift;
    APU.TimerCycles[timer] &= mask;

    // Update the count
    APU.Timer[timer] += updates;
    int numOverflows = APU.Timer[timer] / APU.TimerTarget[timer];
    if (numOverflows > 0) {
        APU.Timer[timer] -= APU.TimerTarget[timer] * numOverflows;
        IAPU.RAM[APU_COUNTER0 + timer] = (IAPU.RAM[APU_COUNTER0 + timer] + numOverflows) & 0xf;

        APU.TimerTarget[timer] = IAPU.RAM[APU_TIMER0 + timer];
        if (APU.TimerTarget[timer] == 0) APU.TimerTarget[timer] = 0x100;
    }
/*    bool hit = false;
    while (APU.Timer[timer] >= APU.TimerTarget[timer]) {
        hit = true;
        APU.Timer[timer] -= APU.TimerTarget[timer];
        IAPU.RAM[APU_COUNTER0 + timer] = (IAPU.RAM[APU_COUNTER0 + timer] + 1) & 0xf;
    }
    if (hit) {
        // Read the new value of the timer target in case it has changed
        APU.TimerTarget[timer] = IAPU.RAM[APU_TIMER0 + timer];
        if (APU.TimerTarget[timer] == 0) APU.TimerTarget[timer] = 0x100;
    }*/
}

void timers9xApu(int cycles) {
    if (APU.TimerEnabled[0])
        updateTimer9x(0, cycles);
    if (APU.TimerEnabled[1])
        updateTimer9x(1, cycles);
    if (APU.TimerEnabled[2])
        updateTimer9x(2, cycles);
}

int last9xCycles = 0;

void step9xApu() {
    APU.Cycles = 0;
    APU_EXECUTE1();
    timers9xApu(APU.Cycles);
    last9xCycles = APU.Cycles;
    APU.Cycles = 0;
}

bool compareState(u8 *mem, u8 *dsp, u8 *extra, u16 pc, char *debugState) {
    bool good = true;
    char *messup = "";
    char buf[100];

    if (last9xCycles != SPC_IPC->lastApuCycles) { good = false; sprintf(buf, "Cycles %d, %d", SPC_IPC->lastApuCycles, last9xCycles); messup = buf; }
    if (APURegisters.YA.B.A != SPC_IPC->apuA) { good = false; messup = "A"; }
    if (APURegisters.X != SPC_IPC->apuX) { good = false; messup = "X"; }
    if (APURegisters.YA.B.Y != SPC_IPC->apuY) { good = false; messup = "Y"; }
    if ((0x100 | APURegisters.S) != SPC_IPC->APU_SP) { good = false; messup = "SP"; }
    S9xAPUPackStatus();
    if (APURegisters.P != SPC_IPC->apuFlags) { good = false; messup = "flags"; }
    if (SPC_IPC->apuPc != (u32)(IAPU.PC - IAPU.RAM)) { good = false; sprintf(buf, "PC %x, %x", pc, (u32)(IAPU.PC - IAPU.RAM)); messup = buf; }

    if (!good) {
        char buf[150], tbuf[150], tbuf2[50];
        S9xDebugGetSPCState(buf);
        DebugGetSPCState(tbuf);
        DebugGetOpcodeDesc(tbuf2, &(mem[pc]), true, mem);
        sprintf(debugState, "Cpu State differs: %s\n%s\n%s\n%s",messup,tbuf2,tbuf,buf);
        return false;
    }

    u32 *ptr1 = (u32*)mem;
    u32 *ptr2 = (u32*)IAPU.RAM;
    for (int i = 0; i < 0x10000 / 4; i++) {
        if (ptr1[i] != ptr2[i]) {
            for (int j = 0; j < 4; j++) {
                if (mem[(i*4)+j] != IAPU.RAM[(i*4)+j]) {
                    char buf[150], tbuf[150], tbuf2[50];
                    S9xDebugGetSPCState(buf);
                    DebugGetSPCState(tbuf);
                    DebugGetOpcodeDesc(tbuf2, &(IAPU.RAM[pc]), true, IAPU.RAM);
                    sprintf(debugState, "Mem differs at %x: %x-%x\n%s\n%s\n%s",i*4+j,mem[i*4+j],IAPU.RAM[i*4+j],tbuf2,tbuf,buf);
                    return false;
                }
            }
        }
    }
/*
    for (int i=0; i<=0x7f; i++) {
        if (dsp[i] != APU.DSP[i]) {
            char buf[150], tbuf[150], tbuf2[50];
            S9xDebugGetSPCState(buf);
            DebugGetSPCState(tbuf);
            DebugGetOpcodeDesc(tbuf2, &(IAPU.RAM[pc]), true, IAPU.RAM);
            sprintf(debugState, "DSP differs at %x: %x-%x\n%s\n%s\n%s",i,dsp[i],APU.DSP[i],tbuf2,tbuf,buf);
            return false;
        }
    }*/

    for (int i=0; i<=0x3f; i++) {
        if (APU.ExtraRAM[i] != extra[i]) {
            char buf[150], tbuf[150], tbuf2[50];
            S9xDebugGetSPCState(buf);
            DebugGetSPCState(tbuf);
            DebugGetOpcodeDesc(tbuf2, &(IAPU.RAM[pc]), true, IAPU.RAM);
            sprintf(debugState, "Extra mem differs at %x: %x-%x\n%s\n%s\n%s",i,extra[i],APU.ExtraRAM[i],tbuf2,tbuf,buf);
            return false;
        }
    }

//    for (int i = 0; i < 8; i++) {
//        if (s9xChannels[0].samplePos != ((DspChannel*)SPC_IPC->channelDebug)->samplePos)

    return true;
}


void s9xExecute(int cycles) {
    APU.Cycles -= cycles;
    do {
        APU.Cycles += S9xAPUCycles [*IAPU.PC];
        (*S9xApuOpcodes[*IAPU.PC])();
    } while (APU.Cycles < 0);
    timers9xApu(cycles);
}

void s9xTestInstruction(u8 a, u8 x, u8 y, u8 psw, u8 sp, u16 pc, u8 op1, u8 op2, u8 op3) {
    IAPU.PC = IAPU.RAM + pc;
    IAPU.DirectPage = IAPU.RAM;
    if (psw & 0x20) IAPU.DirectPage += 0x100;
 
    APURegisters.YA.B.A = a;
    APURegisters.YA.B.Y = y;
    APURegisters.X = x;
    APURegisters.S = sp;
    APURegisters.P = psw;
    S9xAPUUnpackStatus ();
    APURegisters.PC = pc;

    IAPU.RAM[pc] = op1;
    IAPU.RAM[pc+1] = op2;
    IAPU.RAM[pc+2] = op3;

    APU.Cycles = S9xAPUCycles [*IAPU.PC];
    (*S9xApuOpcodes[*IAPU.PC])();
    last9xCycles = APU.Cycles;
}

#endif
